from csocket cimport sockaddr, timeval
from libc.stdint cimport int64_t

cdef extern from "<pcap/pcap.h>" nogil:
    """
    #include <stdio.h>

    // Definitions and stubs for supporting older libpcap versions
    #ifndef PCAP_CHAR_ENC_LOCAL
    #define PCAP_CHAR_ENC_LOCAL	0x00000000U
    #endif

    #ifndef PCAP_CHAR_ENC_UTF_8
    #define PCAP_CHAR_ENC_UTF_8	0x00000001U
    #endif

    #ifndef PCAP_TSTAMP_HOST_HIPREC_UNSYNCED
    #define PCAP_TSTAMP_HOST_HIPREC_UNSYNCED 5
    #endif

    #ifndef PCAP_IF_WIRELESS
    #define PCAP_IF_WIRELESS 0x00000008
    #endif

    #ifndef PCAP_IF_CONNECTION_STATUS
    #define PCAP_IF_CONNECTION_STATUS 0x00000030
    #endif

    #ifndef PCAP_IF_CONNECTION_STATUS_UNKNOWN
    #define PCAP_IF_CONNECTION_STATUS_UNKNOWN 0x00000000
    #endif

    #ifndef PCAP_IF_CONNECTION_STATUS_CONNECTED
    #define PCAP_IF_CONNECTION_STATUS_CONNECTED 0x00000010
    #endif

    #ifndef PCAP_IF_CONNECTION_STATUS_DISCONNECTED
    #define PCAP_IF_CONNECTION_STATUS_DISCONNECTED 0x00000020
    #endif

    #ifndef PCAP_IF_CONNECTION_STATUS_NOT_APPLICABLE
    #define PCAP_IF_CONNECTION_STATUS_NOT_APPLICABLE 0x00000030
    #endif

    #ifndef HAVE_PCAP_INIT
    static int pcap_init(unsigned int opts, char *errbuf)
    {
        return 0;
    }
    #endif

    #ifndef HAVE_DATALINK_VAL_TO_DESCRIPTION_OR_DLT
    static const char * pcap_datalink_val_to_description_or_dlt(int dlt)
    {
            static char unkbuf[40];
            const char *description;

            description = pcap_datalink_val_to_description(dlt);
            if (description != NULL) {
                    return description;
            } else {
                    (void)snprintf(unkbuf, sizeof(unkbuf), "DLT %u", dlt);
                    return unkbuf;
            }
    }
    #endif
    """

    ctypedef unsigned int bpf_u_int32

    enum:
        DLT_NULL
        DLT_EN10MB
        DLT_EN3MB
        DLT_AX25
        DLT_PRONET
        DLT_CHAOS
        DLT_IEEE802
        DLT_ARCNET
        DLT_SLIP
        DLT_PPP
        DLT_FDDI

        DLT_RAW

        DLT_IEEE802_11_RADIO

        DLT_DOCSIS

    struct bpf_insn:
        unsigned short	code
        unsigned char 	jt
        unsigned char 	jf
        bpf_u_int32     k

    struct bpf_program:
        unsigned int bf_len
        bpf_insn* bf_insns

    enum:
        PCAP_ERRBUF_SIZE

    enum:
        PCAP_CHAR_ENC_LOCAL
        PCAP_CHAR_ENC_UTF_8

    struct pcap:
        pass

    ctypedef pcap pcap_t

    struct pcap_dumper:
        pass

    ctypedef pcap_dumper pcap_dumper_t

    ctypedef enum pcap_direction_t:
        PCAP_D_INOUT
        PCAP_D_IN
        PCAP_D_OUT

    struct pcap_pkthdr:
        timeval ts
        bpf_u_int32 caplen
        bpf_u_int32 len

    struct pcap_stat:
        unsigned int ps_recv
        unsigned int ps_drop
        unsigned int ps_ifdrop

    struct pcap_if:
        pcap_if* next
        char* name
        char* description
        pcap_addr* addresses
        bpf_u_int32 flags

    enum:
        PCAP_IF_LOOPBACK
        PCAP_IF_UP
        PCAP_IF_RUNNING
        PCAP_IF_WIRELESS
        PCAP_IF_CONNECTION_STATUS
        PCAP_IF_CONNECTION_STATUS_UNKNOWN
        PCAP_IF_CONNECTION_STATUS_CONNECTED
        PCAP_IF_CONNECTION_STATUS_DISCONNECTED
        PCAP_IF_CONNECTION_STATUS_NOT_APPLICABLE

    ctypedef pcap_if pcap_if_t

    struct pcap_addr:
        pcap_addr* next
        sockaddr* addr
        sockaddr* netmask
        sockaddr* broadaddr
        sockaddr* dstaddr

    ctypedef void (*pcap_handler)(unsigned char *, const pcap_pkthdr *, const unsigned char *) except *

    enum:
        PCAP_ERROR
        PCAP_ERROR_BREAK
        PCAP_ERROR_NOT_ACTIVATED
        PCAP_ERROR_ACTIVATED
        PCAP_ERROR_NO_SUCH_DEVICE
        PCAP_ERROR_RFMON_NOTSUP
        PCAP_ERROR_NOT_RFMON
        PCAP_ERROR_PERM_DENIED
        PCAP_ERROR_IFACE_NOT_UP
        PCAP_ERROR_CANTSET_TSTAMP_TYPE
        PCAP_ERROR_PROMISC_PERM_DENIED
        PCAP_ERROR_TSTAMP_PRECISION_NOTSUP

    enum:
        PCAP_WARNING
        PCAP_WARNING_PROMISC_NOTSUP
        PCAP_WARNING_TSTAMP_TYPE_NOTSUP

    enum: PCAP_NETMASK_UNKNOWN

    int pcap_init(unsigned int, char *)

    int pcap_lookupnet(const char *, bpf_u_int32 *, bpf_u_int32 *, char *)

    pcap_t *pcap_create(const char *, char *)

    int pcap_set_snaplen(pcap_t *, int)

    int pcap_set_promisc(pcap_t *, bint)

    int pcap_can_set_rfmon(pcap_t *)

    int pcap_set_rfmon(pcap_t *, bint)

    int pcap_set_timeout(pcap_t *, int)

    int pcap_set_tstamp_type(pcap_t *, int)

    int pcap_set_immediate_mode(pcap_t *, bint)

    int pcap_set_buffer_size(pcap_t *, int)

    int pcap_set_tstamp_precision(pcap_t *, int)

    int pcap_get_tstamp_precision(pcap_t *)

    int pcap_activate(pcap_t *)

    int pcap_list_tstamp_types(pcap_t *, int **)

    void pcap_free_tstamp_types(int *)

    int pcap_tstamp_type_name_to_val(const char *)

    const char *pcap_tstamp_type_val_to_name(int)

    const char *pcap_tstamp_type_val_to_description(int)

    enum:
        PCAP_TSTAMP_HOST
        PCAP_TSTAMP_HOST_LOWPREC
        PCAP_TSTAMP_HOST_HIPREC
        PCAP_TSTAMP_ADAPTER
        PCAP_TSTAMP_ADAPTER_UNSYNCED
        PCAP_TSTAMP_HOST_HIPREC_UNSYNCED

    enum:
        PCAP_TSTAMP_PRECISION_MICRO
        PCAP_TSTAMP_PRECISION_NANO

    pcap_t *pcap_open_live(const char *, int, bint, int, char *)

    pcap_t *pcap_open_dead(int, int)

    pcap_t *pcap_open_dead_with_tstamp_precision(int, int, unsigned int)

    pcap_t *pcap_open_offline_with_tstamp_precision(const char *, unsigned int, char *)

    pcap_t *pcap_open_offline(const char *, char *)

    void pcap_close(pcap_t *)

    int pcap_loop(pcap_t *, int, pcap_handler, unsigned char *) except *

    int pcap_dispatch(pcap_t *, int, pcap_handler, unsigned char *) except *

    const unsigned char *pcap_next(pcap_t *, pcap_pkthdr *)

    int pcap_next_ex(pcap_t *, pcap_pkthdr **, const unsigned char **)

    void pcap_breakloop(pcap_t *)

    int pcap_stats(pcap_t *, pcap_stat *)

    int pcap_setfilter(pcap_t *, bpf_program *)

    int pcap_setdirection(pcap_t *, pcap_direction_t);

    int pcap_getnonblock(pcap_t *, char *)

    int pcap_setnonblock(pcap_t *, bint, char *)

    int pcap_inject(pcap_t *, const void *, size_t)

    int pcap_sendpacket(pcap_t *, const unsigned char *, int)

    const char *pcap_statustostr(int)

    char* pcap_geterr(pcap_t *)

    int	pcap_compile(pcap_t *, bpf_program *, const char *, int, bpf_u_int32)

    void pcap_freecode(bpf_program *)

    bint pcap_offline_filter(const bpf_program *, const pcap_pkthdr *, const unsigned char *)

    int pcap_datalink(pcap_t *)

    int pcap_list_datalinks(pcap_t *, int **)

    int pcap_set_datalink(pcap_t *, int)

    void pcap_free_datalinks(int *)

    int pcap_datalink_name_to_val(const char *)

    const char *pcap_datalink_val_to_name(int)

    const char *pcap_datalink_val_to_description(int)

    const char *pcap_datalink_val_to_description_or_dlt(int)

    int	pcap_snapshot(pcap_t *)

    bint pcap_is_swapped(pcap_t *)

    int pcap_major_version(pcap_t *)

    int pcap_minor_version(pcap_t *)

    pcap_dumper_t *pcap_dump_open(pcap_t *, const char *)

    pcap_dumper_t *pcap_dump_open_append(pcap_t *, const char *)

    long pcap_dump_ftell(pcap_dumper_t *)

    int64_t pcap_dump_ftell64(pcap_dumper_t *)

    int pcap_dump_flush(pcap_dumper_t *)

    void pcap_dump_close(pcap_dumper_t *)

    void pcap_dump(unsigned char *, const pcap_pkthdr *, const unsigned char *)

    int pcap_findalldevs(pcap_if_t **, char *)

    void pcap_freealldevs(pcap_if_t *)

    const char *pcap_lib_version()

    void bpf_dump(const bpf_program *, int)



cimport cpcap

cdef extern from *:
    """
    #ifdef _WIN32
    #include <Windows.h>
    #include <stdio.h>
    #include <wchar.h>
    static int load_npcap_dlls(void)
    {
        WCHAR npcap_dir[512];
        UINT len;
        len = GetSystemDirectoryW(npcap_dir, 480);
        if (!len) {
            PyErr_SetFromWindowsErr(0);
            return -1;
        }

        wcscat_s(npcap_dir, 512, L"\\\\Npcap");
        if (SetDllDirectoryW(npcap_dir) == 0) {
            PyErr_SetFromWindowsErr(0);
            return -1;
        }

        pcap_lib_version();

        if (SetDllDirectoryW(NULL) == 0) {
            PyErr_SetFromWindowsErr(0);
            return -1;
        }

        return 0;
    }
    #else
    static int load_npcap_dlls(void)
    {
        return 0;
    }
    #endif
    """

    int load_npcap_dlls() except -1


load_npcap_dlls()



import sys
import os
import contextlib
import re
from io import open
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from distutils.errors import CompileError, LinkError


USE_CYTHON = os.path.exists(".git")

if USE_CYTHON:
    from Cython.Build import cythonize
else:
    def cythonize(extensions, **_ignore):
        for extension in extensions:
            sources = []
            for sfile in extension.sources:
                path, ext = os.path.splitext(sfile)
                if ext in ('.pyx', '.py'):
                    if extension.language == 'c++':
                        ext = '.cpp'
                    else:
                        ext = '.c'
                    sfile = path + ext
                sources.append(sfile)
            extension.sources[:] = sources
        return extensions


with open("cypcap.pyx", "r", encoding="utf-8") as f:
    version = re.search(r'(?m)^__version__ = u"([a-zA-Z0-9.-]+)"', f.read()).group(1)

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()


def _has_function(compiler, funcname, includes=None, include_dirs=None,
                 libraries=None, library_dirs=None):
    """Return a boolean indicating whether funcname is supported on
    the current platform.  The optional arguments can be used to
    augment the compilation environment.
    """
    # this can't be included at module scope because it tries to
    # import math which might not be available at that point - maybe
    # the necessary logic should just be inlined?
    import tempfile
    if includes is None:
        includes = []
    if include_dirs is None:
        include_dirs = []
    if libraries is None:
        libraries = []
    if library_dirs is None:
        library_dirs = []
    fd, fname = tempfile.mkstemp(".c", "compiletest", text=True)
    f = os.fdopen(fd, "w")
    try:
        for incl in includes:
            f.write("""#include "%s"\n""" % incl)
        f.write("""\
int main (int argc, char **argv) {
%s;
return 0;
}
""" % funcname)
    finally:
        f.close()
    try:
        objects = compiler.compile([fname], include_dirs=include_dirs)
    except CompileError:
        return False
    finally:
        os.remove(fname)

    try:
        compiler.link_executable(objects, "a.out",
                             libraries=libraries,
                             library_dirs=library_dirs)
    except (LinkError, TypeError):
        return False
    else:
        os.remove(compiler.executable_filename("a.out"))
    finally:
        for fn in objects:
            os.remove(fn)

    return True


@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr
    e.g.:
    with stdchannel_redirected(sys.stderr, os.devnull):
        ...
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, 'w')
        os.dup2(dest_file.fileno(), stdchannel.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


def has_function(*args, **kw):
    with stdchannel_redirected(sys.stderr, os.devnull):
        return _has_function(*args, **kw)


if sys.platform == "win32":
    try:
        npcap_sdk = os.environ["NPCAP_SDK"]
    except KeyError:
        # TODO Better message
        print("Please define NPCAP_SDK", file=sys.stderr)
        sys.exit(1)

    include_dirs = [os.path.join(npcap_sdk, "Include")]
    if sys.maxsize > 2**32:
        library_dirs = [os.path.join(npcap_sdk, R"Lib\x64")]
    else:
        library_dirs = [os.path.join(npcap_sdk, "Lib")]
    libraries = ["delayimp", "Ws2_32", "wpcap"]
    extra_link_args = ["/DELAYLOAD:wpcap.dll"]
else:
    include_dirs = []
    library_dirs = []
    libraries = ["pcap"]
    extra_link_args = []


class build_ext(_build_ext):
    def run(self):
        from distutils import ccompiler

        if self.define is None:
            self.define  = []

        compiler = ccompiler.new_compiler()
        if has_function(
            compiler,
            "pcap_init(0, NULL)",
            includes=["pcap/pcap.h"],
            include_dirs=include_dirs,
            libraries=libraries,
            library_dirs=library_dirs):
                self.define.append(("HAVE_PCAP_INIT", 1))
        if has_function(
            compiler,
            "pcap_datalink_val_to_description_or_dlt(0)",
            includes=["pcap/pcap.h"],
            include_dirs=include_dirs,
            libraries=libraries,
            library_dirs=library_dirs):
                self.define.append(("HAVE_DATALINK_VAL_TO_DESCRIPTION_OR_DLT", 1))

        _build_ext.run(self)


setup(
    name="cypcap",
    version=version,
    author="Segev Finer",
    author_email="segev208@gmail.com",
    description="A Cython based binding for modern libpcap",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/segevfiner/cypcap",
    project_urls={
        "Documentation": "https://segevfiner.github.io/cypcap/",
        "Issue Tracker": "https://github.com/segevfiner/cypcap/issues",
    },
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: System :: Networking :: Monitoring",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="libpcap pcap",
    zip_safe=False,
    ext_modules=cythonize(
        [
            Extension(
                "cypcap", ["cypcap.pyx"],
                include_dirs=include_dirs,
                library_dirs=library_dirs,
                libraries=libraries,
                extra_link_args=extra_link_args,
                # define_macros=[("CYTHON_TRACE_NOGIL", 1)],
            ),
        ],
        # compiler_directives={'linetrace': True},
    ),
    cmdclass={"build_ext": build_ext},
    python_requires='>=3.6',
    extras_require={
        "dev": [
            "cython>=0.29.24",
            "sphinx==4.*",
            "pytest",
            "dpkt",
            "netifaces",
        ],
    },
)



cdef extern from * nogil:
    """
    #ifdef _WIN32
    #include <WinSock2.h>
    #include <WS2tcpip.h>
    #else
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #endif
    """
    ctypedef int socklen_t

    enum:
        AF_INET
        AF_INET6

    enum:
        INET_ADDRSTRLEN
        INET6_ADDRSTRLEN

    struct sockaddr:
        unsigned short sa_family
        char sa_data[14]

    struct in_addr:
        unsigned long s_addr

    struct sockaddr_in:
        unsigned short sin_family
        unsigned short sin_port
        in_addr sin_addr
        char sin_zero[8]

    struct in6_addr:
        unsigned char s6_addr[16]

    struct sockaddr_in6:
        unsigned short sin6_family
        unsigned short sin6_port
        unsigned long sin6_flowinfo
        in6_addr sin6_addr
        unsigned long sin6_scope_id

    const char *inet_ntop(int af, const void* src,
                          char* dst, socklen_t size)

    struct timeval:
        long tv_sec
        long tv_usec
